
Unit TimeDate;

{$I Debug.Inc}

Interface

Type
   DateType = Record
      Year, Month, Day, DOW : Word;
   End; {Record}

   TimeType = Record
      Hour, Minute, Second, MSecond : Word;
   End; {Record}

Procedure SystemTime(Var Time : TimeType);
Procedure SystemDate(Var Date : DateType);
Function LeapYear(Year : Word) : Boolean;
Function ValidDate(Date : DateType) : Boolean;
Function DateToSerial (Date : DateType) : LongInt;
Procedure SerialToDate (SerialDate : LongInt; Var Date : DateType);
Function DateToJulian (Date : DateType) : Word;
Procedure ParseDate (Var InpString; Var Date : DateType);
Procedure SetDayOfWeek (Var Date : DateType);
Function MonthToString(Month : Word) : String;
Procedure GetFirstLast(Date : DateType; Var FirstDay, LastDay, NofDays : Word);
Function StringToMonth (Var MnthString) : Word;
Function StringToDay (Var DyString) : Word;
Procedure Easter(Year : Word; Var Date : DateType);
Function Holiday(Date : DateType): String;
Procedure IncDate(Var Date : DateType; Amount : LongInt);
Function CmpDates(Date1, Date2 : DateType) : LongInt;

Implementation

Const
   DaysInMonth : Array[1..12] of Word =
      (0,31,59,90,120,151,181,212,243,273,304,334);
   MonthArray : Array[1..12] of Word =
      (31,28,31,30,31,30,31,31,30,31,30,31);

{*******************************LEAP YEAR*************************************}
{*                                                                           *}
{*  This function returns true if the year submitted is a leap year.         *}
{*                                                                           *}
{*****************************************************************************}

Function LeapYear(Year : Word) : Boolean;

Begin
   If (Year Mod 4 = 0) or ((Year Div 100) Mod 4 = 0) then
      LeapYear := True
   Else
      LeapYear := False;
End;

{*******************************VALID DATE************************************}
{*                                                                           *}
{*  This function returns true if the date submitted is valid or false if it *}
{*  is not.                                                                  *}
{*                                                                           *}
{*****************************************************************************}

Function ValidDate(Date : DateType) : Boolean;

Begin
   ValidDate := True;
   With Date do
      Begin
         If (Month > 12)  Or (Month < 1) Or (DOW > 6) Or (Day < 1) then
            Begin
               ValidDate := False;
               Exit;
            End;
         If (Month = 2) And (LeapYear(Year) = True) then
            Dec(Day);
         If Day > MonthArray[Month] then
            ValidDate := False;
      End;
End;

{*****************************SET DAY OF WEEK*********************************}
{*                                                                           *}
{*  This procedure returns the day of the week in the input record.          *}
{*                                                                           *}
{*****************************************************************************}

Procedure SetDayOfWeek (Var Date : DateType);

Var
  Century,
  Tmp : Integer;
  TempDate : DateType;

Begin
   TempDate := Date;
   With TempDate do
      Begin
         If Year < 1900 then
            Inc(Year,1900);
         If Month < 3 then
            Inc(Month, 10)
         else
            Dec(Month, 2);
         If Month > 10 then
            Dec(Year);
         Century := Year div 100;
         Year := Year mod 100;
         Tmp := Trunc((2.6 * Month - 0.2) + Day + Year + (Year div 4) +
            (Century div 4) - (2 * Century));
      End;
   Date.DOW := (Tmp + 777) mod 7;
End; {SetDayOfWeek}

{****************************String To Month**********************************}
{*                                                                           *}
{*  This function returns an word value for a character string which         *}
{*  evaluates to one of the 12 months of the year. The first three ordinal   *}
{*  values of the input string are added together and compared to 12 unique  *}
{*  values which correspond to the 12 months. If no comparison is found, 13  *}
{*  becomes the returned value. Case of the input string is of no consequence*}
{*                                                                           *}
{*****************************************************************************}

Function StringToMonth (Var MnthString): Word;

Var
   MonthString : String Absolute MnthString;
   i, MonthVal : Word;

Begin
   MonthVal := 0;
   For i := 1 to 3 do
      MonthVal := MonthVal + (Ord(MonthString[i]) Or $20);
   Case MonthVal of
        313    :  StringToMonth := 1;
        301    :  StringToMonth := 2;
        320    :  StringToMonth := 3;
        323    :  StringToMonth := 4;
        327    :  StringToMonth := 5;
        333    :  StringToMonth := 6;
        331    :  StringToMonth := 7;
        317    :  StringToMonth := 8;
        328    :  StringToMonth := 9;
        326    :  StringToMonth := 10;
        339    :  StringToMonth := 11;
        300    :  StringToMonth := 12;
        Else      StringToMonth := 13;
   End; {Case}
End; {StringToMonth}

{*****************************STRING TO DAY***********************************}
{*                                                                           *}
{*  This function returns a word value for a character string which          *}
{*  evaluates to one of the 7(0..6) days of the week. The first three        *}
{*  ordinal values of the input string are added together and compared to    *}
{*  7 unique values which correspond to the 7 days. If no comparison is      *}
{*  found, 7 becomes the returned value. Case of the input string is of no   *}
{*  consequence                                                              *}
{*                                                                           *}
{*****************************************************************************}

Function StringToDay (Var DyString): Word;

Var
   i, DayVal : Word;
   DayString : String Absolute DyString;

Begin
   DayVal := 0;
   For i := 1 to 3 do
      DayVal := DayVal + (Ord(DayString[i]) Or $20);
   Case DayVal of
      342 : StringToDay := 0;
      330 : StringToDay := 1;
      334 : StringToDay := 2;
      320 : StringToDay := 3;
      337 : StringToDay := 4;
      321 : StringToDay := 5;
      328 : StringToDay := 6;
      Else  StringToDay := 7;
   End; {Case}
End; {String To Day}

{****************************MONTH TO STRING**********************************}
{*                                                                           *}
{*  This function returns the month string of the given ordinal value.       *}
{*                                                                           *}
{*****************************************************************************}

Function MonthToString(Month : Word) : String;

Const
   MonthStArray : Array[1..12] of String = ('January', 'February', 'March',
      'April', 'May', 'June', 'July', 'August', 'September', 'October',
      'November', 'December');

Begin
   MonthToString := MonthStArray[Month];
End; {MonthToString}

{*****************************DATE TO JULIAN**********************************}
{*                                                                           *}
{*  This function accepts a date and returns the julian date equivalent.     *}
{*                                                                           *}
{*****************************************************************************}

Function DateToJulian (Date : DateType) : Word;

Begin
   If (Date.Month > 2) And (LeapYear(Date.Year) = True) then
      Inc(Date.Day);
   DateToJulian := DaysInMonth[Date.Month] + Date.Day;
End; {DateToJulian}

{******************************Date To Serial*********************************}
{*                                                                           *}
{*  This function accepts the month day and year as inputs and returns       *}
{*  a Lotus 123-type serial date.                                            *}
{*                                                                           *}
{*****************************************************************************}

Function DateToSerial (Date : DateType): LongInt;

Var
   Sdate : LongInt;

Begin
   If Date.Year > 1900 then
      Date.Year := Date.Year - 1900;
   Sdate := 365 * Date.Year + (Date.Year + 3) div 4;         {365 for every   }
   If (LeapYear(Date.Year) = True) And (Date.Month > 2) then {plus 1 day for  }
      Inc(Sdate);                                            {each leap year  }
   DateToSerial := Sdate + DaysInMonth[Date.Month] + Date.Day;
End; {DateToSerial}


{*******************************Serial To Date********************************}
{*                                                                           *}
{*  This function returns Word values for the month, day and year as         *}
{*  calculated from Lotus 123-type serial dates. This function fixes an      *}
{*  error in Lotus which fails to predict that 1900 is a leap year. All      *}
{*  other leap years are accouted for.                                       *}
{*                                                                           *}
{*****************************************************************************}

Procedure SerialToDate (SerialDate : LongInt; Var Date : DateType);

Var
   Remainder,                                {Intermediate calculation        }
   LeapYears : Real;                         {Number of leap years since 1900 }
   JulianDate : LongInt;                     {Julian number of days           }
   i : Integer;                              {Counter                         }

Begin {SerialToDate}
   Date.Year := Trunc (SerialDate / 365.26);        {First determine the }
   LeapYears := (Date.Year + 3) Div 4;              {year by subtracting }
   Remainder := (Date.Year * 365.0) + LeapYears;    {365 days for each   }
   JulianDate := Round(SerialDate - Remainder);
   If ((LeapYear(Date.Year) = True) and (JulianDate > 59)) then
      Dec(JulianDate);                                   {a leap year adjust 1}
   i := 14;
   Repeat                                                {Locate the correct  }
      Dec(i);                                            {month by comparison }
   Until   (JulianDate > DaysInMonth[i]) or (i = 0);     {with month table    }
   Date.Month := i;
   Date.Day := JulianDate - DaysInMonth [Date.Month];{Set the day   }
   If (LeapYear(Date.Year)= True) And (Date.Month = 2) then
      Inc(Date.Day);                                {add the day back in }
   SetDayOfWeek(Date);                              {we took out earlier }
End; {SerialToDate}


{********************************Parse Date***********************************}
{*                                                                           *}
{*  This function accepts a variety of date strings as input and produces    *}
{*  Word values for the month day and year. The day must be in the range     *}
{*  of 1-31. The month may either be numerical or spelled out. The year can  *}
{*  be complete (e.g. 1955) or abreviated (e.g. 55). If all values are       *}
{*  numerical the first value is presupposed to be the month, otherwise the  *}
{*  correct month is determined. The '-', '/', or a space may be used as     *}
{*  delimiters. If a valid date is returned then ParseDate will be equal to  *}
{*  0. Otherwise an error code of 0ffh will be returned.                     *}
{*                                                                           *}
{*****************************************************************************}

Procedure ParseDate (Var InpString; Var Date : DateType);

Const
   Delimiters : Set of Char = ['/', '-', ' '];
   Numbers : Set of Char = ['1','2','3','4','5','6','7','8','9','0'];

Var
   InputString : String Absolute InpString;
   StringLength,
   FirstCharPos,
   i, j, k,
   MonthTest : Word;
   Parameter   : Array [1..3] of String;
   CharTest : Boolean;

{*****************************String To Num***********************************}
{*                                                                           *}
{*  This sub-function returns an Word value from an input string of          *}
{*  characters. No check is made to see if the characters are in fact        *}
{*  integers so errors in the returned data can result. Each character is    *}
{*  sequentially accessed, converted into its ordinal equivalent and the     *}
{*  accumulated total is multiplied by 10 for correct decimal position       *}
{*                                                                           *}
{*****************************************************************************}

Function StringToNum (Var NumString) : Word;

Var
   NumericalString : String Absolute NumString;
   StringLength,
   i,
   ReturnValue,
   ZeroAsciiValue : Word;

Begin  {StringToNum}
   StringLength := Ord(NumericalString [0]);
   ZeroAsciiValue := Ord('0');
   ReturnValue := 0;
   For i := 1 to StringLength do
      ReturnValue := ReturnValue * 10 + (Ord(NumericalString[i]) -
                     ZeroAsciiValue);
   StringToNum := ReturnValue;
End;  {StringToNum}

{*******************************Main Routine**********************************}

Begin {ParseDate}
   If InputString = '' then
      Exit;
   StringLength := Ord(InputString[0]);
   For i := 1 to 3 do
      Parameter[i] := ' ';
   i := 0;

{*****************************************************************************}
{*                                                                           *}
{*  This section of code finds the position of the first character in the    *}
{*  input string in case the string is padded with leading delimiters        *}
{*                                                                           *}
{*****************************************************************************}

   Repeat
      Inc(i);
      FirstCharPos := i;
   Until Not((i <> StringLength) And (InputString[i] in Delimiters));

{*****************************************************************************}
{*                                                                           *}
{*  This section of code separates the string into 3 sub-strings. Multiple   *}
{*  delimiters between each substring is allowed. Additional data items      *}
{*  after the third are ignored.                                             *}
{*                                                                           *}
{*****************************************************************************}

   j := 1;
   k := 1;
   i := FirstCharPos - 1;
   Repeat
      Inc(i);
      If Not(InputString [i] in Delimiters) then
         Begin
            CharTest := True;
            Parameter [j][k] := InputString [i];
            Parameter [j][0] := Chr (k);
            Inc(k);
         End
      Else
         If CharTest = True then
            Begin
               Inc(j);
               k := 1;
               CharTest := False
            End;
   Until (i = StringLength) or (j > 3);
{End;}

{*****************************************************************************}
{*                                                                           *}
{*  This section of code determines which of the first two sub-strings       *}
{*  represent the month and whether it is alphanumeric or numerical. The     *}
{*  sub-string is then converted to a Word value by the appropriate          *}
{*  sub-function.                                                            *}
{*                                                                           *}
{*****************************************************************************}

   If Parameter [2][1] in Numbers then
      MonthTest := 1
   Else
      MonthTest := 2;
   If Not(Parameter [MonthTest] [1] in Numbers) then
      Date.Month := StringToMonth(Parameter[MonthTest])
   Else
      Date.Month := StringToNum(Parameter [MonthTest]);

{*****************************************************************************}
{*                                                                           *}
{*  The final two sections use the StringToNum sub-function to convert the   *}
{*  day and year sub-strings into Word values. No leap year                  *}
{*  checking is performed or whether the day is appropriate for the month.   *}
{*                                                                           *}
{*****************************************************************************}

   MonthTest := MonthTest Xor 3;
   Date.Day := StringToNum (Parameter [MonthTest]);
   Date.Year := StringToNum (Parameter[3]);
   Date.DOW := 0;
   If Date.Year < 1900 then
      Date.Year := Date.Year + 1900;
   If ValidDate(Date) = True then
      SetDayOfWeek(Date)
   Else
      FillChar(Date, SizeOf(Date), 0);
End; {ParseDate}

{*****************************GET FIRST LAST**********************************}
{*                                                                           *}
{*  This procedure takes as input the month and year as Word values and      *}
{*  returns the first day of the month (sun - sat, 0 - 6), the last day of   *}
{*  of the month and the number of days in the month.  In order to return    *}
{*  the first two variables, SetDayOfWeek is called.                         *}
{*                                                                           *}
{*****************************************************************************}

Procedure GetFirstLast(Date : DateType; Var FirstDay, LastDay, NofDays : Word);

Begin
   Date.Day := 1;
   SetDayofWeek(Date);
   FirstDay := Date.DOW;
   If (LeapYear(Date.Year) = True) and (Date.Month = 2) then
      NofDays := 29
   Else
      NofDays := MonthArray[Date.Month];
   Date.Day := NofDays;
   SetDayofWeek(Date);
   LastDay := Date.DOW;
End; {GetFirstLast}

{*********************************EASTER**************************************}
{*                                                                           *}
{*  This procedure takes as input the year as a Word value and returns the   *}
{*  month and day of Easter in that year.  I won't attempt to describe the   *}
{*  mathematics behind this laborious calculation, but if you're really      *}
{*  interested see Donald Knuth's Fundamental Algorithms for a more detailed *}
{*  explaination.                                                            *}
{*                                                                           *}
{*****************************************************************************}

Procedure Easter(Year : Word; Var Date : DateType);

Var
   GoldenNo,
   Sun,
   Century,
   LeapCent,
   LunarCorr,
   Epact,
   FullMoon : Integer;

Begin
   Date.Year := Year;
   GoldenNo := (Year Mod 19) + 1;
   Century := (Year Div 100) + 1;
   LeapCent := (3 * Century Div 4) - 12;
   LunarCorr := ((8 * Century + 5) Div 25) - 5;
   Sun := (5 * Year Div 4) - LeapCent - 10;
   Epact := Abs(11 * GoldenNo + 20 + LunarCorr - LeapCent) Mod 30;
   If ((Epact = 25) And (GoldenNo > 11)) Or (Epact = 24) then
      Inc(Epact);
   FullMoon := 44 - Epact;
   If FullMoon < 21 then
      Inc(FullMoon, 30);
   Date.Day := FullMoon + 7 - ((Sun + FullMoon) Mod 7);
   If Date.Day > 31 then
      Begin
         Dec(Date.Day, 31);
         Date.Month := 4;
      End
   Else
      Date.Month := 3;
   Date.DOW := 0;
End;

{********************************HOLIDAY**************************************}
{*                                                                           *}
{*  This function returns a string if the input date falls on a holiday.     *}
{*                                                                           *}
{*****************************************************************************}

Function Holiday(Date : DateType): String;

Const
   HolidaySt : Array[1..10] of String = ('New Year`s','M. L. King`s Day',
      'President`s Day','Easter','Memorial Day','4th of July','Labor Day',
      'Columbus Day','Thanksgiving','Christmas');
Type
   IndexType = (First, Second, Third, Fourth, Last);
   DowType = (Sun, Mon, Tue, Wed, Thur, Fri, Sat);

Var
   HolidayArray : Array [1..10] of Word;
   Year : Word;
   TempDate : DateType;
   FirstDay, LastDay, FirstDOW : Word;
   i : Integer;

   Procedure IndexDay(Index : IndexType; SearchDOW : DowType;
                      Var Date : DateType);

   Var
      FirstDay, LastDay, NofDays : Word;

   Begin
      Date.Dow := Ord(SearchDow);
      GetFirstLast(Date, FirstDay, LastDay, NofDays);
      If Index = Last then
         Begin
            Date.Day := NofDays - (LastDay - Ord(SearchDOW));
            Exit;
         End;
      Date.Day := Date.Dow - FirstDay + 1;
      If FirstDay > Ord(SearchDow) then
         Inc(Date.Day, 7);
      Inc(Date.Day, Ord(Index) * 7);
   End;{IndexDay}

   Procedure FillHolidayArray;

   Begin
      TempDate.Year := Date.Year;
      TempDate.Month := 1;
      TempDate.Day := 1;
      HolidayArray[1] := DateToJulian(TempDate);              {New Year's Day}
      IndexDay(Third, Mon, TempDate);
      HolidayArray[2] := DateToJulian(TempDate);              {M. L. King Day}
      TempDate.Month := 2;
      IndexDay(Third, Mon, TempDate);
      HolidayArray[3] := DateToJulian(TempDate);              {President's day}
      Easter(TempDate.Year, TempDate);
      HolidayArray[4] := DateToJulian(TempDate);              {Easter}
      TempDate.Month := 5;
      IndexDay(Last, Mon, TempDate);
      HolidayArray[5] := DateToJulian(TempDate);              {Memorial day}
      TempDate.Month := 7;
      TempDate.Day := 4;
      HolidayArray[6] := DateToJulian(TempDate);              {4th of July}
      TempDate.Month := 9;
      IndexDay(First, Mon, TempDate);
      HolidayArray[7] := DateToJulian(TempDate);              {Labor day}
      TempDate.Month := 10;
      IndexDay(Second, Mon, TempDate);
      HolidayArray[8] := DateToJulian(TempDate);              {Columbus day}
      TempDate.Month := 11;
      IndexDay(Fourth, Thur, TempDate);
      HolidayArray[9] := DateToJulian(TempDate);              {Thanksgiving}
      TempDate.Month := 12;
      TempDate.Day := 25;
      HolidayArray[10] := DateToJulian(TempDate);             {Christmas}
   End;

Begin
   Holiday := '';
   FillHolidayArray;
   For i := 1 to 10 do
      If DateToJulian(Date) = HolidayArray[i] then
         Holiday := HolidaySt[i];
End;

{********************************INC DATE*************************************}
{*                                                                           *}
{*  This function adds a specified number of days to the given date and      *}
{*  returns the date resulting.  Amount can be negative for subtraction.     *}
{*                                                                           *}
{*****************************************************************************}

Procedure IncDate(Var Date : DateType; Amount : LongInt);

Begin
   SerialToDate(DateToSerial(Date) + Amount, Date);
End;{IncDate}

{********************************CMP DATES************************************}
{*                                                                           *}
{*  This function compares two dates and returns the number of days between  *}
{*  then.  If date two is larger a negative number is returned.              *}
{*                                                                           *}
{*****************************************************************************}

Function CmpDates(Date1, Date2 : DateType) : LongInt;

Begin
   CmpDates := DateToSerial(Date1) - DateToSerial(Date2);
End;{CmpDates}

{$L TimeDate.Obj}
Procedure SystemTime(Var Time : TimeType); External;
Procedure SystemDate(Var Date : DateType); External;

Begin
End.



